//
// k e y m a p . c
//
// An educational keyboard mapping program
//
// Nov. 11, 2002 by W.Nishida (http://www.skyfree.org)
//

#include "kbc.h"
#include "dos-console.h"

asm(".code16gcc");

// Global data for keyboard mapping

#define MOD_ALT		8
#define MOD_CTRL	4
#define MOD_SHIFT	2
#define MOD_CAPS	1

unsigned char modifier = 0;	// Status of shift and toggle keys
unsigned char capslock = 0;	// Capital lock state

// System scan codes

#define ESC	0x01
#define CTRL	0x1D
#define KEY_A	0x1E
#define KEY_SQ	0x28
#define SHIFT	0x2A
#define ALT	0x38
#define CAPS	0x3A

// Keyboard mapping data based on each state (JP106)

char map_plain[] = { 'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', ':' };
char map_ctrl[] =  {0x01,0x13,0x04,0x06,0x07,0x08,0x0A,0x0B,0x0C,0x7F,0x7F };
char map_shift[] = { 'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', '+', '*' };
char map_alt[]   = {0x81,0x93,0x84,0x86,0x87,0x88,0x8A,0x8B,0x8C,0xFF,0xFF };

// LED control routine

void ledctrl(int leds) {
  int ret;

  kbc_write(KBC_WRITE_LED);		// Send write LED command to kbd
  ret = kbc_read();			// Read response code from kbd
  if (ret != KBC_ACK) {			// Acknowledged?
    prints("LED command failure!\r\n");
    return;
   }
  kbc_write(leds);			// Turn on or off LEDs
  ret = kbc_read();			// Read response code from kbd
  if (ret != KBC_ACK)			// Acknowledged?
    prints("LED command failure!\r\n");
 }

int main(void) {
  int ret, up;
  char ch, *map;

  ledctrl(0);				// Turn off all LEDs
  prints("Hit any key, please.\r\n");

  // Inhibit scan code translation

  kbc_command(KBC_GET_CTRL);		// Read current 8042 control mode
  ret = kbc_read();
  kbc_command(KBC_SET_CTRL);		// Update 8042 control mode
  ret &= ~KBC_INT_KEY;			// Disable keyboard interrupt
  kbc_write(ret);			// Send new control mode value

  // Read scancodes repeatedly

  while (1) {
    ret = kbc_read();			// Read system scan code
    up = ret & 0x80;			// Get break bit
    ret &= 0x7F;			// Turn off break bit
    switch (ret) {
      case ESC:				// ESC key
        goto abort;			// Abort job

      case ALT:				// ALT key
        if (up)				// Break?
          modifier &= ~MOD_ALT;
        else				// Make
          modifier |= MOD_ALT;
        continue;			// scan again

      case CTRL:			// CTRL key
        if (up)				// Break?
          modifier &= ~MOD_CTRL;
        else				// Make
          modifier |= MOD_CTRL;
        continue;			// scan again

      case SHIFT:			// SHIFT key
        if (up)				// Break?
          modifier &= ~MOD_SHIFT;
        else				// Make
          modifier |= MOD_SHIFT;
        continue;			// scan again

      case CAPS:			// CAPS key
        if (up)	{			// Break?
          modifier &= ~MOD_CAPS;
          capslock ^= 1;		// Invert capital lock state
          ledctrl(capslock << 2);	// Switch CAPS lock LED
         }
        else				// Make
          modifier |= MOD_CAPS;
        continue;			// scan again

      case KEY_A...KEY_SQ:		// ASDFGHJKL;'
        if (up)				// Key was depressed?
          continue;			// Then, ignore it
        else {				// Key was pressed
          if (modifier & MOD_SHIFT)
            map = map_shift;
          else if (modifier & MOD_CTRL)
            map = map_ctrl;
          else if (modifier & MOD_ALT)
            map = map_alt;
          else
            map = map_plain;
         }
        ch = map[ ret - KEY_A ];
        if (capslock) {			// CAPS is locked
          if ((modifier & MOD_SHIFT) == 0) {
            if (ch >= 'a' && ch <= 'z')
              ch -= 0x20;
           }
          else {
            if (ch >= 'A' && ch <= 'Z')
              ch += 0x20;
           }
         }
        break;

      default:				// Unknown key
        if (up == 0) printc('\007');	// Warn to user
        continue;			// Ignore it
     }
    dump_byte(ch);			// Display mapped character code
    printc(' ');
   }

  // Recover scan code translation
 abort:
  kbc_command(KBC_GET_CTRL);		// Read current 8042 control mode
  ret = kbc_read();
  kbc_command(KBC_SET_CTRL);		// Update 8042 control mode
  ret |= KBC_INT_KEY;			// Enable keyboard interrupt
  kbc_write(ret);			// Send new control mode value

  return 0;
 }
